from django.apps import AppConfig


class DjangoProseEditorConfig(AppConfig):
    name = "django_prose_editor"

    def ready(self):
        # Import system checks
        from . import checks  # noqa: F401



from django.apps import apps
from django.conf import settings
from django.core.checks import Error, Warning, register


@register()
def check_js_preset_configuration(app_configs, **kwargs):
    """
    Check that the 'default' JavaScript preset is not being overridden in settings.
    """
    errors = []

    # Main setting
    if hasattr(settings, "DJANGO_PROSE_EDITOR_PRESETS"):
        presets = settings.DJANGO_PROSE_EDITOR_PRESETS
        if "default" in presets:
            errors.append(
                Error(
                    'Overriding the "default" preset in DJANGO_PROSE_EDITOR_PRESETS is not allowed.',
                    hint="Remove the 'default' key from your DJANGO_PROSE_EDITOR_PRESETS setting.",
                    obj=settings,
                    id="django_prose_editor.E001",
                )
            )

        if "configurable" in presets:
            errors.append(
                Error(
                    'Overriding the "configurable" preset in DJANGO_PROSE_EDITOR_PRESETS is not allowed.',
                    hint="Remove the 'configurable' key from your DJANGO_PROSE_EDITOR_PRESETS setting.",
                    obj=settings,
                    id="django_prose_editor.E002",
                )
            )

    return errors


@register()
def check_extensions_parameter(app_configs, **kwargs):
    """
    Check for usage of 'extensions' (because we want that!)
    """
    from django_prose_editor.fields import ProseEditorField

    warnings = []

    # Get models to check based on provided app_configs or all models
    models_to_check = []
    if app_configs:
        for app_config in app_configs:
            models_to_check.extend(app_config.get_models())
    else:
        models_to_check = apps.get_models()

    # Check models for fields using deprecated config
    for model in models_to_check:
        for field in model._meta.fields:
            if isinstance(field, ProseEditorField):
                # Check if this field is using legacy config (no 'extensions' key)
                if (
                    not isinstance(field.config, dict)
                    or "extensions" not in field.config
                ):
                    warnings.append(
                        Warning(
                            "This ProseEditorField is using the legacy configuration format which is "
                            "deprecated and will be removed in a future version. Add the 'extensions' "
                            "configuration explicitly to use the new configuration format.",
                            obj=f"{model._meta.label}.{field.name}",
                            id="django_prose_editor.W001",
                        )
                    )

    return warnings


@register()
def check_custom_extensions_configuration(app_configs, **kwargs):
    """
    Check that custom extensions are properly configured.
    """
    errors = []

    # Check if custom extensions are defined
    if hasattr(settings, "DJANGO_PROSE_EDITOR_EXTENSIONS"):
        extensions = settings.DJANGO_PROSE_EDITOR_EXTENSIONS

        # Check that extensions is a list
        if not isinstance(extensions, list):
            errors.append(
                Error(
                    "DJANGO_PROSE_EDITOR_EXTENSIONS must be a list of dictionaries.",
                    hint="Configure DJANGO_PROSE_EDITOR_EXTENSIONS as a list of dictionaries, each with 'js' and 'extensions' keys.",
                    obj=settings,
                    id="django_prose_editor.E003",
                )
            )
            return errors

        # Check each extension group
        for i, extension_group in enumerate(extensions):
            if not isinstance(extension_group, dict):
                errors.append(
                    Error(
                        f"Extension group at index {i} must be a dictionary.",
                        hint="Each extension group must be a dictionary with 'js' and 'extensions' keys.",
                        obj=settings,
                        id="django_prose_editor.E004",
                    )
                )
                continue

            # Check that required keys are present
            if "extensions" not in extension_group:
                errors.append(
                    Error(
                        f"Extension group at index {i} is missing the required 'extensions' key.",
                        hint="Each extension group must have an 'extensions' key mapping extension names to processors.",
                        obj=settings,
                        id="django_prose_editor.E005",
                    )
                )
                continue

            if "js" not in extension_group:
                errors.append(
                    Warning(
                        f"Extension group at index {i} is missing the 'js' key.",
                        hint="Each extension group should have a 'js' key listing JavaScript assets for the extensions.",
                        obj=settings,
                        id="django_prose_editor.W002",
                    )
                )

            # Check the extensions dictionary
            extensions_dict = extension_group["extensions"]
            if not isinstance(extensions_dict, dict):
                errors.append(
                    Error(
                        f"The 'extensions' key in extension group at index {i} must be a dictionary.",
                        hint="The 'extensions' key should map extension names to processor callables or dotted paths.",
                        obj=settings,
                        id="django_prose_editor.E006",
                    )
                )
                continue

            # Check each processor
            for extension_name, processor in extensions_dict.items():
                # Check if the processor is a string (dotted path) or callable
                if not (callable(processor) or isinstance(processor, str)):
                    errors.append(
                        Error(
                            f'Processor for extension "{extension_name}" in group {i} must be a callable or a dotted path string.',
                            hint="Each processor must be either a callable or a dotted import path.",
                            obj=settings,
                            id="django_prose_editor.E007",
                        )
                    )

                # If it's a string, verify it looks like a dotted path
                if isinstance(processor, str) and "." not in processor:
                    errors.append(
                        Warning(
                            f'Processor path "{processor}" for extension "{extension_name}" in group {i} may not be a valid dotted import path.',
                            hint="The processor should be a dotted import path like 'myapp.processors.my_processor'.",
                            obj=settings,
                            id="django_prose_editor.W003",
                        )
                    )

            # Check the js assets list
            js_assets = extension_group.get("js", [])
            if not isinstance(js_assets, (list, tuple)):
                errors.append(
                    Error(
                        f"The 'js' key in extension group at index {i} must be a list.",
                        hint="The 'js' key should be a list of JavaScript asset URLs.",
                        obj=settings,
                        id="django_prose_editor.E008",
                    )
                )

    return errors


@register()
def check_sanitization_enabled(app_configs, **kwargs):
    """
    Check that all ProseEditorField instances have sanitization enabled.
    """
    from django_prose_editor.fields import ProseEditorField, _actually_empty

    warnings = []

    # Get models to check based on provided app_configs or all models
    models_to_check = []
    if app_configs:
        for app_config in app_configs:
            models_to_check.extend(app_config.get_models())
    else:
        models_to_check = apps.get_models()

    # Check all models for ProseEditorField without sanitization
    for model in models_to_check:
        for field in model._meta.fields:
            if isinstance(field, ProseEditorField):
                # Check if sanitization is disabled or set to the identity function
                if field.sanitize == _actually_empty:
                    # Different messages based on whether using extensions or legacy config
                    if isinstance(field.config, dict) and "extensions" in field.config:
                        message = (
                            "This ProseEditorField is using extensions without sanitization. "
                            "For security, it's recommended to enable sanitization with "
                            "sanitize=True when using extensions."
                        )
                        hint = (
                            "Add sanitize=True to this field definition for proper HTML sanitization "
                            "that matches your configured extensions."
                        )
                    else:
                        message = (
                            "This ProseEditorField doesn't have sanitization enabled. "
                            "For security, it's recommended to enable sanitization."
                        )
                        hint = (
                            "Consider using the newer extensions mechanism with sanitize=True "
                            "for proper HTML sanitization that matches your editor capabilities."
                        )

                    warnings.append(
                        Warning(
                            message,
                            hint=hint,
                            obj=f"{model._meta.label}.{field.name}",
                            id="django_prose_editor.W004",
                        )
                    )

    return warnings



"""
Configuration system for Django Prose Editor.

This module provides a way to define editor extensions and generate
corresponding sanitization rules for server-side HTML cleaning.
"""

from typing import Any

from django.conf import settings
from django.utils.module_loading import import_string


def add_tags_and_attributes(shared_config, tags, attributes):
    # Add tags to the config
    shared_config["tags"].update(tags)

    # Add attributes to the config
    if attributes:
        for tag, allowed_attrs in attributes.items():
            if tag not in shared_config["attributes"]:
                shared_config["attributes"][tag] = set()
            shared_config["attributes"][tag].update(allowed_attrs)


def html_tags(tags, attributes=None):
    """
    Create a simple processor function that adds tags and attributes to the sanitization config.

    Args:
        tags: List of HTML tags
        attributes: Dict mapping tags to allowed attributes

    Returns:
        A processor function that updates the shared config
    """

    def processor(extension_config, shared_config):
        add_tags_and_attributes(shared_config, tags, attributes)

    return processor


def process_heading(config, shared_config):
    """Process Heading extension configuration."""
    levels = (
        c if isinstance(config, dict) and (c := config.get("levels")) else range(1, 7)
    )
    tags = [f"h{level}" for level in levels]
    add_tags_and_attributes(shared_config, tags, None)


def process_link(config, shared_config):
    """Process Link extension configuration."""
    # Prepare tags and attributes
    tags = ["a"]
    attributes = {"a": ["href", "rel", "title"]}
    # If rel is in the list of allowed attributes link_rel has to be set to
    # None.
    shared_config["link_rel"] = None

    # Include target attribute unless explicitly disabled
    target_allowed = True
    if isinstance(config, dict):
        if "enableTarget" in config:
            target_allowed = bool(config["enableTarget"])

        # Include URL schemes if specified (for nh3 sanitization)
        if "protocols" in config:
            if "url_schemes" not in shared_config:
                shared_config["url_schemes"] = set()
            shared_config["url_schemes"].update(config["protocols"])

    # Add target attribute if allowed
    if target_allowed:
        attributes["a"].append("target")

    # Add tags and attributes to the shared config
    add_tags_and_attributes(shared_config, tags, attributes)


# Additional processors for other extensions
def process_code_block(config, shared_config):
    """Process CodeBlock extension configuration."""
    # Prepare tags and attributes
    tags = ["pre", "code"]
    attributes = {}

    # Include language attribute if supported
    if isinstance(config, dict) and config.get("languageClassPrefix"):
        attributes["pre"] = ["class"]
        attributes["code"] = ["class"]

    # Add tags and attributes to the shared config
    add_tags_and_attributes(shared_config, tags, attributes)


def process_figure(config, shared_config):
    """Process Figure extension configuration."""
    # Prepare tags and attributes
    tags = ["figure", "figcaption", "img"]
    attributes = {
        "img": ["src", "alt", "width", "height"],
        "figure": ["class"],
    }

    # Add tags and attributes to the shared config
    add_tags_and_attributes(shared_config, tags, attributes)


def process_image(config, shared_config):
    """Process Image extension configuration."""
    # Prepare tags and attributes
    tags = ["img"]
    attributes = {
        "img": ["src", "alt", "width", "height"],
    }

    # Add tags and attributes to the shared config
    add_tags_and_attributes(shared_config, tags, attributes)


def process_text_align(config, shared_config):
    """Process TextAlign extension configuration."""
    # Text align typically uses classes, not new tags
    tags = []
    # Text align adds style and class attributes to various elements
    attributes = {
        "p": ["style", "class"],
        "h1": ["style", "class"],
        "h2": ["style", "class"],
        "h3": ["style", "class"],
        "h4": ["style", "class"],
        "h5": ["style", "class"],
        "h6": ["style", "class"],
        "blockquote": ["style", "class"],
    }

    # Add tags and attributes to the shared config
    add_tags_and_attributes(shared_config, tags, attributes)


def process_color_highlight(config, shared_config):
    """Process Color and Highlight extension configuration."""
    # Prepare tags and attributes
    tags = ["span"]
    attributes = {
        "span": ["style", "class"],
    }

    # Add tags and attributes to the shared config
    add_tags_and_attributes(shared_config, tags, attributes)


# Default extension mapping with processors as the single source of truth
EXTENSION_MAPPING = {
    # Core formatting
    "Bold": html_tags(["strong"]),
    "Italic": html_tags(["em"]),
    "Strike": html_tags(["s"]),
    "Underline": html_tags(["u"]),
    "Subscript": html_tags(["sub"]),
    "Superscript": html_tags(["sup"]),
    # Code extensions
    "Code": html_tags(["code"]),
    "CodeBlock": process_code_block,
    # Text styling
    "Color": process_color_highlight,
    "Highlight": process_color_highlight,
    "TextAlign": process_text_align,
    "TextStyle": process_color_highlight,
    # Structure
    "Heading": process_heading,
    "Paragraph": html_tags(["p"]),
    "HardBreak": html_tags(["br"]),
    "BulletList": html_tags(["ul"]),
    "OrderedList": html_tags(["ol"], {"ol": ["start", "type"]}),
    "ListItem": html_tags(["li"]),
    "Blockquote": html_tags(["blockquote"]),
    "HorizontalRule": html_tags(["hr"]),
    # Media and figures
    "Image": process_image,
    "Figure": process_figure,
    "Caption": html_tags(["figcaption"]),
    # Advanced extensions
    "Link": process_link,
    "Table": html_tags(
        ["table", "tbody", "thead", "tr", "th", "td"],
        {
            "th": ["rowspan", "colspan"],
            "td": ["rowspan", "colspan"],
        },
    ),
    "TableRow": html_tags([]),
    "TableHeader": html_tags([]),
    "TableCell": html_tags([]),
    # Special extensions (these don't produce HTML elements)
    "History": html_tags([]),
    "HTML": html_tags([]),
    "Typographic": html_tags([]),
    "Document": html_tags([]),
    "Text": html_tags([]),
    "Dropcursor": html_tags([]),
    "Gapcursor": html_tags([]),
    "Menu": html_tags([]),
    "NoSpellCheck": html_tags([]),
}

# Automatic dependencies (extensions that require other extensions)
EXTENSION_DEPENDENCIES = {
    "BulletList": ["ListItem"],
    "OrderedList": ["ListItem"],
    "Table": ["TableRow", "TableHeader", "TableCell"],
    "Figure": ["Caption", "Image"],
    "TextAlign": ["TextStyle"],
    "Color": ["TextStyle"],
    "Highlight": ["TextStyle"],
}


def expand_extensions(extensions: dict[str, Any]) -> dict[str, Any]:
    """
    Expand extension configuration by applying defaults and resolving dependencies.

    Args:
        extensions: Dictionary of extension configurations

    Returns:
        Expanded extension configuration
    """

    expanded = {
        # Core extensions
        "Document": True,
        "Dropcursor": True,
        "Gapcursor": True,
        "Paragraph": True,
        "Text": True,
        "Menu": True,
        "NoSpellCheck": True,
        # Enable history by default unless explicitly disabled
        "History": True,
    }

    # Include user-specified extensions
    expanded.update(extensions)

    # Filter out disabled extensions
    expanded = {
        extension: config
        for extension, config in expanded.items()
        if config is not False
    }

    # Resolve dependencies
    for extension, deps in EXTENSION_DEPENDENCIES.items():
        if extension in expanded:
            for dep in deps:
                if dep not in expanded:
                    expanded[dep] = True

    return expanded


def js_from_extensions(
    extensions: dict[str, Any],
) -> list[str]:
    """
    Get JavaScript modules required by the enabled extensions.

    Args:
        extensions: Dictionary of extension configurations

    Returns:
        List of JavaScript module paths
    """
    expanded = set(expand_extensions(extensions))
    js_modules = set()

    extensions = getattr(settings, "DJANGO_PROSE_EDITOR_EXTENSIONS", [])
    for group in extensions:
        if (js := group.get("js")) and expanded & group["extensions"].keys():
            js_modules.update(js)

    return list(js_modules)


def allowlist_from_extensions(
    extensions: dict[str, Any],
) -> dict[str, Any]:
    """
    Generate HTML sanitization config from extension configuration.

    This function collects configuration from all extension processors into a unified
    configuration dictionary suitable for nh3 HTML sanitization.

    Each processor function should accept two arguments:
    1. extension_config: The extension-specific configuration
    2. shared_config: The shared configuration dictionary to update

    Processors should directly modify the shared_config dictionary with their
    required HTML elements, attributes, and other nh3 parameters.

    Args:
        extensions: Dictionary of extension configurations

    Returns:
        Dictionary with sanitization config
    """
    expanded = expand_extensions(extensions)

    # Initialize the combined configuration with empty sets
    nh3_config = {
        "tags": set(),
        "attributes": {},
    }

    extensions = getattr(settings, "DJANGO_PROSE_EDITOR_EXTENSIONS", [])
    processors = EXTENSION_MAPPING.copy()
    for group in extensions:
        processors |= group["extensions"]

    for extension, config in expanded.items():
        processor = processors[extension]
        if isinstance(processor, str):
            processors[extension] = import_string(extension)
        processor(config, nh3_config)

    return nh3_config



from copy import deepcopy

from django_prose_editor.fields import ProseEditorField, _actually_empty


def _nh3_sanitizer():
    import nh3

    attributes = deepcopy(nh3.ALLOWED_ATTRIBUTES)
    attributes["a"].add("target")
    attributes["ol"] |= {"start", "type"}

    return lambda x: _actually_empty(nh3.clean(x, attributes=attributes))


class SanitizedProseEditorField(ProseEditorField):
    def __init__(self, *args, **kwargs):
        if "sanitize" not in kwargs:
            kwargs["sanitize"] = _nh3_sanitizer()
        super().__init__(*args, **kwargs)



import re

from django import forms
from django.contrib.admin import widgets
from django.db import models
from django.utils.html import strip_tags
from django.utils.text import Truncator

from django_prose_editor.config import (
    allowlist_from_extensions,
    expand_extensions,
)
from django_prose_editor.widgets import AdminProseEditorWidget, ProseEditorWidget


def _actually_empty(x):
    """
    ProseMirror's schema always adds at least one empty paragraph

    We want empty fields to actually be empty strings so that those field
    values evaluate as ``False`` in a boolean context.
    """
    if re.match(r"^<(?P<tag>\w+)></(?P=tag)>$", x):
        return ""
    return x


def _identity(x):
    return x


def create_sanitizer(extensions):
    """Create a sanitizer function based on extension configuration."""
    try:
        import nh3
    except ImportError:
        raise ImportError(
            "You need to install nh3 to use automatic sanitization. "
            "Install django-prose-editor[sanitize] or pip install nh3"
        )

    nh3_kwargs = allowlist_from_extensions(expand_extensions(extensions))
    return lambda html: _actually_empty(nh3.clean(html, **nh3_kwargs))


def _create_sanitizer(argument, config):
    if argument is False:
        return _actually_empty

    if argument is True:
        return create_sanitizer(config["extensions"])

    if isinstance(argument, (list, tuple)):
        argument = [
            fn(config["extensions"]) if fn == create_sanitizer else fn
            for fn in argument
        ]

        def apply(html):
            for fn in reversed(argument):
                html = fn(html)
            return html

        return apply

    return argument


class ProseEditorField(models.TextField):
    """
    The field has two modes: Legacy mode and normal mode. Normal mode is
    activated by passing an ``config`` dict containing an ``extensions`` key.
    This mode is described below. See the README for the legacy mode.

    A field that uses a unified configuration for both editor extensions and sanitization.

    This field automatically synchronizes the editor capabilities with server-side
    sanitization rules, ensuring that what users can create in the editor matches
    what is allowed after sanitization.

    Args:
        config: Dictionary mapping extension names to their configuration
        preset: Optional JavaScript preset name to override the default
        sanitize: Whether to enable sanitization or a custom sanitizer function
    """

    def __init__(self, *args, **kwargs):
        self.config = kwargs.pop("config", {})
        if extensions := kwargs.pop("extensions", None):
            self.config["extensions"] = extensions

        if "extensions" in self.config:
            # Normal mode
            self.sanitize = _create_sanitizer(
                kwargs.pop("sanitize", False), self.config
            )
            self.preset = kwargs.pop("preset", "configurable")

        else:
            # Legacy mode
            self.sanitize = _create_sanitizer(kwargs.pop("sanitize", False), None)
            self.preset = kwargs.pop("preset", "default")

        super().__init__(*args, **kwargs)

    def clean(self, value, instance):
        return self.sanitize(super().clean(value, instance))

    def contribute_to_class(self, cls, name, **kwargs):
        """Add a ``get_*_excerpt`` method to models which returns a
        de-HTML-ified excerpt of the contents of this field"""
        super().contribute_to_class(cls, name, **kwargs)
        setattr(
            cls,
            f"get_{name}_excerpt",
            lambda self, words=10, truncate=" ...": Truncator(
                strip_tags(getattr(self, name))
            ).words(words, truncate=truncate),
        )

    def deconstruct(self):
        name, _path, args, kwargs = super().deconstruct()
        return (name, "django.db.models.TextField", args, kwargs)

    def formfield(self, **kwargs):
        defaults = {
            "config": self.config,
            "form_class": ProseEditorFormField,
            "preset": self.preset,
        } | kwargs
        return super().formfield(**defaults)


def _is(widget, widget_class):
    return (
        issubclass(widget, widget_class)
        if isinstance(widget, type)
        else isinstance(widget, widget_class)
    )


class ProseEditorFormField(forms.CharField):
    widget = ProseEditorWidget

    def __init__(self, *args, **kwargs):
        self.config = kwargs.pop("config", {})
        if extensions := kwargs.pop("extensions", None):
            self.config["extensions"] = extensions

        if "extensions" in self.config:
            # Normal mode
            self.sanitize = _create_sanitizer(
                kwargs.pop("sanitize", _identity), self.config
            )
            self.preset = kwargs.pop("preset", "configurable")

        else:
            # Legacy mode
            self.sanitize = kwargs.pop("sanitize", _identity)
            self.preset = kwargs.pop("preset", "default")

        widget = kwargs.get("widget")

        # We don't know if widget is set, and if it is, we do not know if it is
        # a class or an instance of the widget. The following if statement
        # should take all possibilities into account.
        if widget and _is(widget, widgets.AdminTextareaWidget):
            kwargs["widget"] = AdminProseEditorWidget
        elif not widget or not _is(widget, ProseEditorWidget):
            kwargs["widget"] = ProseEditorWidget

        super().__init__(*args, **kwargs)
        self.widget.config = self.config
        self.widget.preset = self.preset

    def clean(self, value):
        return self.sanitize(super().clean(value))
